<?php
/**
 * @file
 * Contains helper classes for the SDL BeGlobal plugin specific operations
 * depending on the translation quality selected by the user.
 */

/**
 * Interface TMGMTSDLBeGlobalTranslatorPluginHelperInterface
 */
interface TMGMTSDLBeGlobalTranslatorPluginHelperInterface {
  function doPost($parameters = array(), $tokens = array(), $action);
  function doGet($parameters = array(), $tokens = array(), $action);
  function doPut($parameters = array(), $tokens = array(), $action);
  function doDelete($parameters = array(), $tokens = array(), $action);
  function doRequestTranslation(TMGMTJob $job);
}

/**
 * Class TMGMTSDLBeGlobalTranslatorPluginHelperAbstract
 */
abstract class TMGMTSDLBeGlobalTranslatorPluginHelperAbstract implements TMGMTSDLBeGlobalTranslatorPluginHelperInterface {

  /**
   * The plugin controller which uses this helper.
   *
   * @var TMGMTSDLBeGlobalTranslatorPluginController
   */
  protected $plugin_controller;

  /**
   * The translator object.
   *
   * @var TMGMTTranslator
   */
  protected $translator;

  /**
   * The action for which this helper is required. It is a uri form a list of
   * available actions validated by the controller.
   *
   * @var string
   */
  private $action;

  /**
   * Constructs a new helper for the translation plugin.
   *
   * @param TMGMTSDLBeGlobalTranslatorPluginController $plugin_controller
   * @param TMGMTTranslator $translator
   * @param string $action
   */
  public function __construct(TMGMTSDLBeGlobalTranslatorPluginController $plugin_controller,
                              TMGMTTranslator $translator, $action) {

    $this->plugin_controller = $plugin_controller;
    $this->translator = $translator;
    $this->action = $action;
  }

  /**
   * Creates a POST request to SDL BeGlobal Translate service.
   *
   * @param array $parameters
   *   (Optional) Additional data parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   * @param string $action
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doPost($parameters = array(), $tokens = array(), $action = '') {
    if (!empty($action)) {
      $this->action = $action;
    }
    $response = drupal_http_request(
      url($this->plugin_controller->getTranslationUri() . $this->replaceActionTokens($tokens)),
      array(
        'data' => json_encode($parameters),
        'method' => 'POST',
        'headers' => array(
          'Content-type' => 'application/json',
          'Authorization' => 'BeGlobal apiKey=' . $this->plugin_controller->api_key,
        ),
      )
    );

    $this->validateResponse($response);

    // Return the processed JSON result.
    return drupal_json_decode($response->data);
  }

  /**
   * Creates a POST request with a file to SDL BeGlobal Translate service.
   *
   * @param array $parameters
   *   (Optional) Additional data parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   * @param string $action
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doFilePost($parameters = array(), $tokens = array(), $action = '') {
    if (!empty($action)) {
      $this->action = $action;
    }
    $boundary = 'A0sFSD';
    $response = drupal_http_request(
      url($this->plugin_controller->getTranslationUri() . $this->replaceActionTokens($tokens)),
      array(
        'data' => $this->multipartEncode($boundary, $parameters),
        'method' => 'POST',
        'headers' => array(
          'Content-type' => 'multipart/form-data; boundary=' . $boundary,
          'Authorization' => 'BeGlobal apiKey=' . $this->plugin_controller->api_key,
        ),
      )
    );

    $this->validateResponse($response);

    // Return the processed JSON result.
    return drupal_json_decode($response->data);
  }

  /**
   * Creates a GET request to SDL BeGlobal Translate service.
   *
   * @param array $parameters
   *   (Optional) Additional query parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   * @param string $action
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doGet($parameters = array(), $tokens = array(), $action = '') {
    if (!empty($action)) {
      $this->action = $action;
    }
    $url = url(
      $this->plugin_controller->getTranslationUri() . '/' . $this->replaceActionTokens($tokens),
      array('query' => $parameters)
    );
    $options['headers']['Authorization'] = 'BeGlobal apiKey=' . $this->plugin_controller->api_key;

    $response = drupal_http_request($url, $options);

    $this->validateResponse($response);

    // Return the processed JSON result.
    return drupal_json_decode($response->data);
  }

  /**
   * Creates a PUT request to SDL BeGlobal Translate service.
   *
   * @param array $parameters
   *   (Optional) Additional query parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   * @param string $action
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doPut($parameters = array(), $tokens = array(), $action = '') {
    if (!empty($action)) {
      $this->action = $action;
    }
    $response = drupal_http_request(
      url($this->plugin_controller->getTranslationUri() . '/' . $this->replaceActionTokens($tokens)),
      array(
        'data' => json_encode($parameters),
        'method' => 'PUT',
        'headers' => array(
          'Content-type' => 'application/json',
          'Authorization' => 'BeGlobal apiKey=' . $this->plugin_controller->api_key,
        ),
      )
    );

    $this->validateResponse($response);

    // Return the processed JSON result.
    return drupal_json_decode($response->data);
  }

  /**
   * Creates a DELETE request to SDL BeGlobal Translate service.
   *
   * @param array $parameters
   *   (Optional) Additional query parameters to be passed into the request.
   * @param array $tokens
   *   (Optional) Tokens to be replaced in the action url pattern.
   * @param string $action
   *
   * @return array
   *   Decoded JSON response from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   *   - Unable to connect to SDL BeGlobal Translate service
   *   - Error returned by SDL BeGlobal Translate service
   */
  public function doDelete($parameters = array(), $tokens = array(), $action = '') {
    if (!empty($action)) {
      $this->action = $action;
    }
    $url = url($this->plugin_controller->getTranslationUri() . '/' . $this->replaceActionTokens($tokens));
    $options['headers']['Authorization'] = 'BeGlobal apiKey=' . $this->plugin_controller->api_key;

    $response = drupal_http_request($url, $options);
    kpr($response);
    $this->validateResponse($response);

    // Return the processed JSON result.
    return drupal_json_decode($response->data);
  }

  /**
   * Helper function to deal with data processing in order to make a translation
   * request. It has to be overridden to suit the needs of each type of
   * translation quality.
   *
   * @param TMGMTJob $job
   *
   * @throws TMGMTSDLBeGlobalException
   *   In case of rejected translation.
   */
  public function doRequestTranslation(TMGMTJob $job) {
    // Needs to be overridden.
  }

  /**
   * Fetches the remote mapping of the job from our side.
   *
   * @param TMGMTJob $job
   * @return bool|mixed
   */
  public function getJobRemoteMapping(TMGMTJob $job) {
    // We return FALSE here because if we really have a remote mapping we will
    // use it in a specific way when we override this method.
    return FALSE;
  }

  /**
   * Validates the response received from SDL BeGlobal Translate service and
   * checks for thrown errors.
   *
   * @param object $response
   *   The response object from SDL BeGlobal Translate service.
   *
   * @throws TMGMTSDLBeGlobalException
   */
  protected function validateResponse($response) {
    if ($response->code != 200) {
      throw new TMGMTSDLBeGlobalException(
        t('Unable to connect to SDL BeGlobal Translate service due to following error: @error.'),
        array('@error' => $response->error)
      );
    }
    // If we got error.
    if (!empty($response->error)) {
      throw new TMGMTSDLBeGlobalException(
        t('SDL BeGlobal Translate service returned following error: @error'),
        array('@error' => $response->error)
      );
    }
  }

  /**
   * Checks if there are tokens in the url and take the replacement tokens from
   * the parameters.
   *
   * @param array $tokens
   *   An array containing as key the parameter (projectId, fileId, jobId) and
   *   as value the respective ID received from SDL BeGlobal.
   *
   * @return string
   *   The action with the tokens replaced
   */
  protected function replaceActionTokens($tokens = array()) {
    if (empty($tokens)) {
      return $this->action;
    }

    $this->action = str_replace(array_keys($tokens), array_values($tokens), $this->action);

    return $this->action;
  }

  /**
   * Helper function to process the parameters used for the file upload.
   *
   * @param string $boundary
   * @param array $parameters
   *
   * @return string
   */
  private function multipartEncode($boundary, $parameters) {
    $output = "--$boundary\r\n";
    foreach ($parameters as $key => $value){
      if ($key == 'file') {
        if (substr($value, 0, 1) == "@") {
          $value = substr($value, 1);
        }
        $filename = basename($value);
        $output .= "Content-Disposition: form-data; name=\"file\"; filename=\"$filename\"\r\n" .
          "Content-Transfer-Encoding: binary\r\n" .
          "Content-Type: application/octet-stream\r\n\r\n" .
          file_get_contents($value) . "\r\n";
      }
      else {
        $output .= "Content-Disposition: form-data; name=\"$key\"\r\n\r\n$value\r\n";
      }
      $output .= "--$boundary";
    }
    $output .= "--$boundary--";

    return $output;
  }
}

class TMGMTSDLBeGlobalTranslatorPluginHelperMachineTranslation extends TMGMTSDLBeGlobalTranslatorPluginHelperAbstract{

  /**
   * {@inheritdoc}
   */
  protected function validateResponse($response) {
    parent::validateResponse($response);
    // If we do not have data.
    if (empty($response->data)) {
      throw new TMGMTSDLBeGlobalException(t('SDL BeGlobal Translate service returned no data'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doRequestTranslation(TMGMTJob $job) {
    // Pull the source data array through the job and flatten it.
    $data = array_filter(tmgmt_flatten_data($job->getData()), '_tmgmt_filter_data');

    $translation = array();
    $e = 0;

    foreach ($data as $key => $value) {
      try {
        $result = $this->doPost(array(
          'from' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->source_language),
          'to' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->target_language),
          'text' => $value['#text'],
        ));
        if (empty($result['translation'])) {
          throw new TMGMTSDLBeGlobalException('Empty translation response string.');
        }
        $translation[$key]['#text'] = $result['translation'];
      }
      catch (TMGMTSDLBeGlobalException $e) {
        $job->addMessage('There was a translation error: !error for translation of text "!text."',
          array('!error' => $e->getMessage(), '!text' => $data[$key]['#text']), 'error');
        $e++;
      }
    }

    // The translation job has been successfully submitted.
    $job->submitted('The translation job @job_name has been submitted.', array('@job_name' => $job->label));

    if ($e != count($data)) {
      // Save the translated data through the job.
      $job->addTranslatedData(tmgmt_unflatten_data($translation));
    }
    else {
      $job->rejected('All items of this job were not translated');
    }
  }
}

class TMGMTSDLBeGlobalTranslatorPluginHelperHumanTranslation extends TMGMTSDLBeGlobalTranslatorPluginHelperAbstract {

  /**
   * {@inheritdoc}
   */
  protected function validateResponse($response) {
    parent::validateResponse($response);
  }

  /**
   * {@inheritdoc}
   */
  public function doRequestTranslation(TMGMTJob $job) {
    $job_items = $job->getItems();
    if (!empty($job_items)) {
      $project = $this->doPost(array(
        'type' => 'translationProject',
        'from' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->source_language),
        'to' => $this->plugin_controller->mapToRemoteLanguage($job->getTranslator(), $job->target_language),
        'name' => $job->label,
        'description' => $job->label,
        'notificationCallbackURL' => url('admin/tmgmt_sdl_beglobal_human_callback', array('absolute' => TRUE)),
      ), array(), 'projects');



      /**
       * - each job item will be put in a .txt file
       * - these files will become files attached to the SDL project
       * - the response from the file creation will be mapped to a remote
       * mapping entity TMGMTRemote which will contain key items from the
       * response, as fileId, projectId in order to be able to process it later
       * when the callback is accessed for notification
       * - place the order for the project
       */
      try {
        foreach ($job_items as $item) {
          $data = array_filter(tmgmt_flatten_data($item->data), '_tmgmt_filter_data');
          $file_path = 'public://sdl_beglobal/job_item_' . $item->tjiid . '.txt';
          // Upload a generated file for this job item.
          $file = $this->uploadFile($file_path, $data, $project);
          dsm($file);
          // Create the remote mapping entity for the current job item.
          $remote_mapping = $item->addRemoteMapping(
            serialize(array_keys($data)),
            $project['project']['id'],
            array(
              'remote_identifier_2' => $file['storedFiles'][0]['id'],
              'remote_data' => serialize($file),
            )
          );
          if (!$remote_mapping) {
            throw new TMGMTSDLBeGlobalException(
              'Remote mapping could not be created for job item @item_id',
              array('@item_id' => $item->tjiid)
            );
          }
        }

        // Make the order to SDL BeGlobal.
        $project = $this->doPut(
          array('qualityLevel' => $this->plugin_controller->getTranslationQuality()),
          array('projectId' => $project['project']['id']),
          '/projects/projectId/orders'
        );

        // Save a remote mapping for the job itself which corresponds to the sdl
        // beglobal project.
        $job = $this->addJobRemoteMapping($job, array(
          'remote_identifier_1' => $project['project']['id'],
          'remote_data' => serialize($project),
        ));
        $job->submitted('Job @job_name has been submitted to translator.', array('@job_name' => $job->label));
      }
      catch (TMGMTSDLBeGlobalException $e) {
        // Reject the current job.
        $job->rejected('Translation has been rejected with following error: !error',
          array('!error' => $e->getMessage()), 'error');
        // Delete the newly created project.
        $this->doDelete(array(), array('projectId' => $project['project']['id']), '/projects/projectId');
        watchdog('TMGMT Warning', $e->getMessage(), WATCHDOG_WARNING);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getJobRemoteMapping(TMGMTJob $job) {
    if (empty($job->reference)) {
      return FALSE;
    }

    return entity_load_single('tmgmt_remote', $job->reference);
  }

  /**
   * Uses the data received through the notification callback URL to download
   * the translation files and add the translations to the entities.
   *
   * - Download the files for the respective job
   * - Update each job item with the received translation
   *
   * @param $projectData
   */
  public function translateJob($projectData) {
    // TODO to be used on the notofication callback from SDL BeGlobal.
    if (!empty($projectData['jobs'])) {
      if (!empty($projectData['project']['id'])) {
        $q = new EntityFieldQuery();
        // Find the remote job mapping.
        $tr = $q->entityCondition('entity_type', 'tmgmt_remote')
          ->propertyCondition('remote_identifier_1', $projectData['project']['id'])
          ->propertyOrderBy('remote_identifier_2')
          ->execute();
        if (!empty($tr)) {
          $remote_mappings = entity_load('tmgmt_remote', array_keys($tr['tmgmt_remote']));
          foreach ($remote_mappings as $job_item_remote) {
            if (!empty($job_item_remote->remote_identifier_2)) {
              $remote_job_items[$job_item_remote->remote_identifier_2] = $job_item_remote;
            }
          }
        }
      }
      if (!empty($remote_job_items)) {
        foreach ($projectData['jobs'] as $job) {
          if (!empty($job['translationFileId'])) {
            $response = $this->doGet(
              array(),
              array('storedFileId' => $job['translationFileId']),
              'files/storedFileId/file'
            );
            if (!empty($response->data)) {
              $data = str_getcsv($response->data, ';', '"');
              if (!empty($data)) {
                $job_item = $remote_job_items[$job['sourceFileId']]->getJobItem();
                $translation = array();
                $keys = unserialize($remote_job_items[$job['sourceFileId']]->data_item_key);
                // Map the translation data to the previously saved keys.
                foreach ($keys as $i => $value) {
                  if (!empty($data[$i])) {
                    $translation[$value]['#text'] = $data[$i];
                  }
                }
                // We add the translated data on the entity to be reviewed and
                // change the status of the job item to needs review.
                $job_item->addTranslatedData($translation);
              }
            }
          }
        }
      }
    }
  }

  /**
   * Uploads a file to SDL BeGlobal and attaches it to a project if provided.
   *
   * @param string $file_path
   * @param array $data
   * @param array $project
   *
   * @return array
   *
   * @throws TMGMTSDLBeGlobalException
   *   - In case of empty data object received from SDL BeGlobal.
   */
  private function uploadFile($file_path, $data, $project = array()) {
    if (empty($data)) {
      throw new TMGMTSDLBeGlobalException(
        'Empty data object received for file @file',
        array('@file' => $file_path)
      );
    }
    $dir_name = dirname($file_path);
    if (file_prepare_directory($dir_name, FILE_CREATE_DIRECTORY)) {
      $fp = fopen($file_path, 'a+');
      $array_values = array();
      foreach ($data as $value) {
        $array_values[] = $value['#text'];
      }
      fputcsv($fp, $array_values, ';', '"');
      fclose($fp);
      $parameters['file'] = file_create_url($file_path);
      if (!empty($project)) {
        $parameters += array(
          'projectId' => $project['project']['id'],
          'token' => $project['uploadToken'],
        );
      }
    }
    else {
      throw new TMGMTSDLBeGlobalException(
        'Unable to create directory or directory is not writeable @dir',
        array('@file' => $dir_name)
      );
    }

    return $this->doFilePost($parameters, array(), 'files');
  }

  /**
   * Creates a remote mapping entity for the remote sdl beglobal project.
   *
   * @param TMGMTJob $job
   * @param array $mapping_data
   *
   * @return mixed
   *
   * @throws TMGMTSDLBeGlobalException
   */
  private function addJobRemoteMapping(TMGMTJob $job, $mapping_data = array()) {
    if (empty($mapping_data['remote_identifier_1'])) {
      throw new TMGMTSDLBeGlobalException('Cannot create remote mapping without remote identifier.');
    }

    // We map only the job on this remote mapper.
    $data = array(
      'tjid' => $job->tjid,
      'tjiid' => 0,
    );

    if (!empty($mapping_data)) {
      $data += $mapping_data;
    }

    $remote_mapping = entity_create('tmgmt_remote', $data);
    if (entity_get_controller('tmgmt_remote')->save($remote_mapping)) {
      $job->reference = $remote_mapping->trid;
    }

    return $job;
  }
}

