<?php

/**
 * @file
 * Provides SDL BeGlobal Translator plugin controller.
 *
 * Check @link https://www.beglobal.com/developers/api-documentation/
 * SDL BeGlobal Cloud Translator @endlink.
 */

/**
 * SDL BeGlobal translator plugin controller.
 */
class TMGMTSDLBeGlobalTranslatorPluginController extends TMGMTDefaultTranslatorPluginController {

  const TRANSLATOR_URL = 'https://api.beglobal.com/';

  /**
   * API key used for authentication of the API calls.
   *
   * @var
   */
  public $api_key;

  /**
   * Translation service URL.
   *
   * @var string
   */
  protected $translatorUrl = self::TRANSLATOR_URL;

  /**
   * Available actions for SDL BeGlobal translator.
   *
   * @var array
   */
  protected $availableActions = array(
    'translate',
    'languages',
    'detect',
    'mt-translation/resourceIDs',
    'projects',
    'projects/projectId/files',
    'projects/projectId/files/fileIds',
    'projects/projectId',
    'files',
    'files/storedFileId',
    'files/storedFileId/file',
    'projects/projectId/orders',
    'projects/projectId/quotes',
    'jobs/jobId',
    'jobs/jobId/accept',
    'jobs/jobId/refine',
    'jobs/jobId/comments',
  );

  /**
   * The translation quality used for API calls.
   *
   * @var string
   */
  protected $translation_quality = MACHINE_TRANSLATION_QUALITY;

  /**
   * Builds an array of options containing the possible types of translation
   * quality.
   *
   * @return array
   */
  public static function getTranslationQualityOptions() {
    return array(
      MACHINE_TRANSLATION_QUALITY => t('Machine translation'),
      STANDARD_QUALITY => t('Standard translation'),
      VERIFIED_QUALITY => t('Verified translation'),
      SPECIALIZED_QUALITY => t('Professional translation'),
    );
  }

  /**
   * Translation URI getter.
   *
   * @return string
   */
  public function getTranslationUri() {
    return $this->translatorUrl;
  }

  /**
   * Available actions getter.
   *
   * @return array
   */
  public function getAvailableActions() {
    return $this->availableActions;
  }

  /**
   * Translation quality getter.
   *
   * @return string
   */
  public function getTranslationQuality() {
    return $this->translation_quality;
  }

  /**
   * Overrides TMGMTDefaultTranslatorPluginController::isAvailable().
   */
  public function isAvailable(TMGMTTranslator $translator) {
    if ($translator->getSetting('api_key')) {

      return TRUE;
    }

    return FALSE;
  }

  /**
   * Implements TMGMTTranslatorPluginControllerInterface::requestTranslation().
   */
  public function requestTranslation(TMGMTJob $job) {
    $this->getHelper($job->getTranslator(), 'translate')->doRequestTranslation($job);
  }

  /**
   * Overrides TMGMTDefaultTranslatorPluginController::getSupportedTargetLanguages().
   */
  public function getSupportedTargetLanguages(TMGMTTranslator $translator, $source_language) {
    $languages = array();
    // Request only the language codes with source language as  $source_language
    // and the quality selected by the user.
    $response = $this->getHelper($translator, 'languages')->doGet(array(
      'from' => $source_language,
      'quality' => $this->translation_quality
    ));

    if (!empty($response['languageExpertise'][$this->translation_quality])) {
      foreach ($response['languageExpertise'][$this->translation_quality] as $pair) {
        $languages[$pair['languagePair']['to']['code']] = $pair['languagePair']['to']['code'];
      }
    }

    return $languages;
  }

  /**
   * Overrides TMGMTDefaultTranslatorPluginController::getSupportedLanguagePairs().
   */
  public function getSupportedLanguagePairs(TMGMTTranslator $translator) {
    $language_pairs = array();
    // Request only the language codes with the quality selected by the user.
    $response = $this->getHelper($translator, 'languages')->doGet(array(
      'quality' => $this->translation_quality
    ));

    if (!empty($response['languageExpertise'][$this->translation_quality])) {
      foreach ($response['languageExpertise'][$this->translation_quality] as $pair) {
        $language_pairs += array(
          'source_language' => $pair['languagePair']['from']['code'],
          'target_language' => $pair['languagePair']['to']['code'],
        );
      }
    }

    return $language_pairs;
  }

  /**
   * Overrides TMGMTDefaultTranslatorPluginController::getSupportedRemoteLanguages().
   */
  public function getSupportedRemoteLanguages(TMGMTTranslator $translator) {
    $languages = array();
    // First time we go to the configuration page for the plugin we get an error
    // because the request will be unauthorized.
    if ($translator->getSetting('api_key')) {
      $response = $this->getHelper($translator, 'languages')->doGet(array(
        'quality' => $this->translation_quality
      ));

      if (!empty($response['languageExpertise'][$this->translation_quality])) {
        foreach ($response['languageExpertise'][$this->translation_quality] as $pair) {
          $languages[$pair['languagePair']['to']['code']] = $pair['languagePair']['to']['code'];
        }
      }
    }

    return $languages;
  }

  /**
   * Overrides TMGMTDefaultTranslatorPluginController::hasCheckoutSettings().
   */
  public function hasCheckoutSettings(TMGMTJob $job) {
    return FALSE;
  }

  /**
   * Implements TMGMTTranslatorPluginControllerInterface::canTranslate().
   */
  public function canTranslate(TMGMTTranslator $translator, TMGMTJob $job) {
    // If we have defined mappings for the local to remote languages we need to
    // take it into account here.
    $target_language = $this->mapToRemoteLanguage($translator, $job->target_language);
    $source_language = $this->mapToRemoteLanguage($translator, $job->source_language);

    // The job is only translatable if the translator is available and the
    // target language is among the supported ones.
    return ($this->isAvailable($translator) &&
      array_key_exists($target_language, $translator->getSupportedTargetLanguages($source_language)));
  }

  /**
   * Cancels a remote project from SDL BeGlobal.
   *
   * @param TMGMTJob $job
   */
  public function deleteRemoteJob(TMGMTJob $job) {
    $helper = $this->getHelper($job->getTranslator(), 'projects/projectId');
    $remote_mapping = $helper->getJobRemoteMapping($job);
    // We should be in this case only if we are on human translation.
    if (!empty($remote_mapping)) {
      $helper->doDelete(array(), array('projectId' => $remote_mapping->remote_identifier_1));
    }
  }

  /**
   * Attempts to cancel a translation process.
   *
   * @param TMGMTJob $job
   *
   * @return bool|void
   */
  public function cancelTranslation(TMGMTJob $job) {
    $this->deleteRemoteJob($job);
    return parent::cancelTranslation($job);
  }

  /**
   * Translation quality setter.
   *
   * @param string $translation_quality
   *
   * @return mixed
   */
  protected function setTranslationQuality($translation_quality) {
    $this->translation_quality = $translation_quality;
  }

  /**
   * Prepare the information for the request to be send to SDL BeGlobal
   * Translate service.
   *
   * @param TMGMTTranslator $translator
   * @param string $action
   *
   * @throws TMGMTSDLBeGlobalException
   */
  protected function prepareRequestInformation(TMGMTTranslator $translator, $action) {
    if (!in_array($action, $this->getAvailableActions())) {
      throw new TMGMTSDLBeGlobalException(
        t('Invalid action requested: @action'), array('@action' => $action));
    }

    // We set the translation quality and the api key before we do the request because in the
    // constructor we do not have access to the translator.
    $this->api_key = $translator->getSetting('api_key');
    if (empty($this->api_key)) {
      throw new TMGMTSDLBeGlobalException(
        t('SDL BeGlobal Translate service requires an API key. Please configure
        your this setting in the configuration page of the translator.')
      );
    }

    $this->setTranslationQuality($translator->getSetting('translation_quality'));
    if (empty($this->translation_quality)) {
      $this->setTranslationQuality(MACHINE_TRANSLATION_QUALITY);
    }
  }

  /**
   * Returns a helper object which deals with the CRUD tasks and also particular
   * tasks dependant on the translation quality.
   *
   * @param TMGMTTranslator $translator
   * @param string $action
   *
   * @return TMGMTSDLBeGlobalTranslatorPluginHelperHumanTranslation|TMGMTSDLBeGlobalTranslatorPluginHelperMachineTranslation
   */
  private function getHelper(TMGMTTranslator $translator, $action) {
    $this->prepareRequestInformation($translator, $action);
    $class = 'TMGMTSDLBeGlobalTranslatorPluginHelperMachineTranslation';

    if (!empty($this->translation_quality) &&
      $this->translation_quality != MACHINE_TRANSLATION_QUALITY) {

      $class = 'TMGMTSDLBeGlobalTranslatorPluginHelperHumanTranslation';
    }

    return new $class($this, $translator, $action);
  }

  /**
   * We provide translatorUrl setter so that we can override its value
   * in automated testing.
   *
   * @param $translator_url
   */
  final function setTranslatorURL($translator_url) {
    $this->translatorUrl = $translator_url;
  }
}
